/**
    This is the main source file of program METACUBE, version 1.0 dated from 2015-06-24.
    Program METACUBE and its source files are licensed under the CC-BY-SA 3.0 Unported License. Namely, you are free to share and to remix this work, under the conditions of attributing the work to its author and distributing the work resulting of an alteration under a compatible license.
    Program METACUBE was written by R&eacute;mi Peyre (Universit&eacute; de Lorraine - CNRS)
    For documentation or reporting bugs, see http://www.phare.normalesup.org/~rpeyre/pro/research/cubetransport/metacube
**/

#include "metacube.h"

int main (int argc, char *argv[])
{


  /** VARIABLE DECLARATIONS **/

  /* Constants */

  char const *const jobname = "metacube";

  /* Important variables */

  char inname[FILENAME_MAX], backname[FILENAME_MAX], outname[FILENAME_MAX]; // Names of resp. the input, backup and ouput files
  FILE *instream, *backstream, *outstream; // Streams corresponding to resp. the input, backup and output files
  int k; // Block size for meta-cubing
  struct int3 nnni, nnno; // Grid size for input file, resp. for output file
  double ***densityi, ***densityo; // Input, resp. output densities

  /* Less important variables */

  bool testi, testk, testo, testp, testv, testw; // Tests for whether the input file, the blocksize argument, etc., were given
  char jobname_old[BUFSIZ]; // Job used to get the input file
  int natoms; // Number of atoms
  struct double3 origi, origo; // Origin of grid for input file, resp. for output file
  struct double33 latti, latto; // Lattice dimensions for input, resp. for output

  /* Auxiliary variables */

  int i; // Generic index
  int x, y, z; // Indices for locating a node in the grid
  int xi, yi, zi; // Indices for initial positions
  int dx, dy, dz; // Indices for position differences
  char *theptr; // Auxiliary pointer to char, used to read input
  char auxline[BUFSIZ]; // Auxiliary string to handle lines of text
  char thechar; // Auxiliary char variable
  int theint; // Auxiliary int variable


  /** READING INPUT **/

  /* Default parameters */

  k = 1;
  logstream = NULL; // Actually no log will be written by this program

  /* Initializing which command arguments were given */

  testi = false;
  testk = false;
  testo = false;
  testp = false;
  testv = false;
  testw = false;
  for (i = 1; i < argc; ++i) // Starting to read the input arguments
    {
      theptr = argv[i];
      if (!strncmp (theptr, "-", 1)) // Case the input argument is for an option
        {
          theptr += 1;

          /* Case the argument is to claim punctual data */

          if (!strcmp (theptr, "p"))
            {
              if (testv)
                {
                  error ("Option \"-p\" conflicts with option \"-v\".\n");
                }
              testp = true;
            }

          /* Case the argument is to claim volumic data */

          else if (!strcmp (theptr, "v"))
            {
              if (testp)
                {
                  error ("Option \"-v\" conflicts with option \"-p\".\n");
                }
              testv = true;
            }

          /* Case the argument is to demand overwriting */

          else if (!strcmp (theptr, "w"))
            {
              if (testo)
                {
                  error ("Option \"-w\" conflicts with option \"-o=\".\n");
                }
              testw = true;
            }

          /* Reading an argument for specifying meta-cubing block size */

          else if (!strncmp (theptr, "k=", 2))
            {
              if (testk)
                {
                  error ("You have given several meta-cubing arguments.\n");
                }
              testk = true;
              theptr += 2;
              theint = sscanf (theptr, "%d", &k);
              if (theint < 1)
                {
                  error ("I could not read your meta-cubing argument \"%s\".\n", theptr);
                }
            }

          /* Reading an argument for specifying the output file name */

          else if (!strncmp (theptr, "o=", 2))
            {
              if (testo)
                {
                  error ("You have given several output arguments.\n");
                }
              if (testw)
                {
                  error ("Option \"-o=\" conflicts with option \"-w\".\n");
                }
              testo = true;
              theptr += 2;
              strncpy_0 (outname, theptr, FILENAME_MAX);
            }
        }

      /* Reading an argument for specifying the input file */

      else
        {
          if (testi)
            {
              error ("You have given several input files.\n");
            }
          strncpy_0 (inname, theptr, FILENAME_MAX);
          testi = true;
        }
    }


  /** COPING WITH THE GIVEN ARGUMENTS **/

  /* Looking if the main argument is missing */

  if (!testi)
    {
      error ("Missing input file.\n");
    }

  /* Creating a backup copy if overwriting was demanded */

  if (testw)
    {
      strncpy_0 (backname, inname, FILENAME_MAX);
      strncat_0 (backname, "~", FILENAME_MAX);
      instream = fopen (inname, "r");
      backstream = fopen (backname, "w");
      while (true)
        {
          thechar = getc (instream);
          if (thechar == EOF)
            {
              break;
            }
          putc (thechar, backstream);
        }
      fclose (backstream);
      fclose (instream);
      strncpy_0 (outname, inname, FILENAME_MAX); // Setting output file to be the same as input
      strncpy_0 (inname, backname, FILENAME_MAX); // Redirecting backup file as input file
    }

  /* Fixing an output filename if none was given */

  else if (!testo)
    {
      strncpy_0 (outname, inname, FILENAME_MAX);
      strncat_0 (outname, "_meta", FILENAME_MAX);
    }


  /** HANDLING HEADERS **/

  /* Opening files */

  instream = fopen_ctrl (inname, "r");
  outstream = fopen_ctrl (outname, "w");

  /* Handling first two lines */

  fgets (auxline, BUFSIZ, instream);
  sscanf (auxline, "%s", jobname_old);
  if (!strcmp (jobname_old, jobname))
    {
      if (testp)
        {
          warning ("You have declared your density data follows a punctual description; yet I see that it was apparently generated by \"%s\", which normally outputs a volumic description: this is likely a mistake from your side.\n", jobname);
        }
      else
        {
          testv = true;
        }
    }
  fprintf (outstream, "%s ", jobname);
  fprintf (outstream, "%s ", testv ? "-v" : "-p");
  fprintf (outstream, "-k=%d\n", k);
  fgets (auxline, BUFSIZ, instream);
  fprintf (outstream, "Reorganized data from file \"%s\"\n", inname);

  /* Reading number of atoms, size of grid and dimensions of the input lattice */

  fgets (auxline, BUFSIZ, instream);
  sscanf (auxline, "%d%lf%lf%lf", &natoms, &origi.x, &origi.y, &origi.z);
  fgets (auxline, BUFSIZ, instream);
  sscanf (auxline, "%d%lf%lf%lf", &nnni.x, &latti.x.x, &latti.x.y, &latti.x.z);
  fgets (auxline, BUFSIZ, instream);
  sscanf (auxline, "%d%lf%lf%lf", &nnni.y, &latti.y.x, &latti.y.y, &latti.y.z);
  fgets (auxline, BUFSIZ, instream);
  sscanf (auxline, "%d%lf%lf%lf", &nnni.z, &latti.z.x, &latti.z.y, &latti.z.z);

  /* Computing size of grid and dimensions of the output lattice */

  nnno.x = (nnni.x - 1) / k + 1;
  nnno.y = (nnni.y - 1) / k + 1;
  nnno.z = (nnni.z - 1) / k + 1;
  origo = origi;
  origo = vectsum (origo, vectmult ((k - 1.) / 2, latti.x));
  origo = vectsum (origo, vectmult ((k - 1.) / 2, latti.y));
  origo = vectsum (origo, vectmult ((k - 1.) / 2, latti.z));
  latto.x = vectmult (k, latti.x);
  latto.y = vectmult (k, latti.y);
  latto.z = vectmult (k, latti.z);

  /* Writing down size of grid and dimensions of the output lattice */

  fprintf (outstream, "%5d\t%12.6f\t%12.6f\t%12.6f\n", natoms, origo.x, origo.y, origo.z);
  fprintf (outstream, "%5d\t%12.6f\t%12.6f\t%12.6f\n", nnno.x, latto.x.x, latto.x.y, latto.x.z);
  fprintf (outstream, "%5d\t%12.6f\t%12.6f\t%12.6f\n", nnno.y, latto.y.x, latto.y.y, latto.y.z);
  fprintf (outstream, "%5d\t%12.6f\t%12.6f\t%12.6f\n", nnno.z, latto.z.x, latto.z.y, latto.z.z);

  /* Copying lines relative to the atomic nuclei */

  for (i = 0; i < natoms; ++i)
    {
      fgets (auxline, BUFSIZ, instream);
      fputs (auxline, outstream);
    }


  /** READING INPUT DENSITIES **/

  printf ("Reading data...\n");
  METAMALLOC (densityi, nnni);
  METAFOR (x, y, z, nnni)
  {
    fscanf (instream, "%lf", &densityi[x][y][z]);
  }
  ENDMETAFOR
  fclose (instream);


  /** INTERPOLATING PUNCTUAL DENSITIES INTO VOLUMIC DENSITIES, IF NEEDED **/

  printf ("Processing data...\n");
  if (!testv)
    {
      recube (&densityi, nnni);
    }


  /** META-CUBING STEP **/

  METAMALLOC (densityo, nnno);
  METAFOR (x, y, z, nnno) // Scanning the cells of "densityo"
  {
    densityo[x][y][z] = 0;
    for (dx = 0; dx < k; ++dx) // Scanning the cells of "densityi" involved in densityo[x][y][z]
      {
        for (dy = 0; dy < k; ++dy)
          {
            for (dz = 0; dz < k; ++dz)
              {
                xi = k * x + dx;
                yi = k * y + dy;
                zi = k * z + dz;
                if (xi < nnni.x && yi < nnni.y && zi < nnni.z)
                  {
                    densityo[x][y][z] += densityi[xi][yi][zi];
                  }
              }
          }
      }
    densityo[x][y][z] /= k * k * k;
  }
  ENDMETAFOR
  METAFREE (nnni, densityi);


  /** WRITING DATA ONTO THE NEW FILE **/

  printf ("Writing data...\n");
  for (x = 0; x < nnno.x; ++x)
    {
      for (y = 0; y < nnno.y; ++y)
        {
          for (z = 0; z < nnno.z; ++z)
            {
              fprintf (outstream, "%13.5e", densityo[x][y][z]);
              if ((z + 1) % 6 == 0 || z + 1 >= nnno.z)
                {
                  fprintf (outstream, "\n");
                }
              else
                {
                  fprintf (outstream, "\t");
                }
            }
        }
    }
  fclose (outstream);
  printf ("Done.\n");
  METAFREE (nnno, densityo);
  return EXIT_SUCCESS;
}
